/*
----------------------
Be Sample Code License
----------------------

Copyright 1991-2001, Be Incorporated.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions, and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions, and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    
*/
#include <Application.h>
#include <Entry.h>
#include <Errors.h>
#include <File.h>
#include <Message.h>
#include <Messenger.h>
#include <netdb.h>
#include <NodeInfo.h>
#include <Path.h>
#include <String.h>
#include <stdlib.h>
#include <time.h>
#include "netcentral.h"

#define DASPORTAL_HELLO 'dhlo'
#define DASPORTAL_CONNECTED 'dcnt'
#define DASPORTAL_HELLO_REPLY 'dhlr'
#define DASPORTAL_GOODBYE 'dbye'
#define UPDATE_QUEUE 'updq'
#define FILE_INFO 'finf'
#define FILE_ABORT 'abrt'
#define FILE_TRY_LATER 'tl8r'
#define DASPORTAL_RESTART 1
#define DASPORTAL_RECEIVER_DONE 'done'

extern BMessage *SettingsMsg;

NetCentral *ME;

NetCentral::NetCentral(const char *name):BLooper(name, B_LOW_PRIORITY, B_LOOPER_PORT_DEFAULT_CAPACITY) {
	#if DEBUG
	PRINT(("NetCentral constructor\n"));
	#endif
	ME = this;
	MainWindow = NULL;
	mConstructor_sem = create_sem(1, "netcentral_const_sem");
	acquire_sem(mConstructor_sem);
	mThreadList_sem = create_sem(1, "threadlist_sem");
	mQueueList_sem = create_sem(1, "queuelist_sem");
	mFS_sem = create_sem(5, "fssem");
	mFR_sem = create_sem(5, "frsem");
	fscount = frcount = 0;
	mBroadcast_ep = new BNetEndpoint(SOCK_DGRAM);
	quit = 0;
	//Setup broadcasting endpoint
	int option = 1;
	mBroadcast_ep->SetOption(SO_BROADCAST, SOL_SOCKET, &option, sizeof( int ));
	mBroadcast_ep->SetNonBlocking(true);
	mBroadcast_ep->SetTimeout(0x186a0);
	if (mBroadcast_ep->InitCheck() != B_OK) {
		#if DEBUG
		PRINT(("Broadcast endpoint is no good. Quitting.\n"));
		#endif
		be_app->PostMessage(B_QUIT_REQUESTED);
		release_sem(mConstructor_sem);
		delete_sem(mConstructor_sem);
		return;
	}
 	//create and start receiver thread.
 	mReceive_th = spawn_thread(SpawnReceiverThread, "receiver_thread", B_LOW_PRIORITY, this);
 	mFRC_th=  spawn_thread(SpawnFRConnection, "connection_thread", B_LOW_PRIORITY, this);
 	resume_thread(mFRC_th);
	if (mReceive_th < 0) {
		#if DEBUG
		BString error;
		switch(mReceive_th) {
			case B_NO_MORE_THREADS:
				error = "No more threads available.";
				break;
			case B_NO_MEMORY:
				error = "Not enough memory available.";
				break;
			default:
				error = "Unknown error.";
		}
		PRINT(("Couldn't set up receiver thread: %s. Quitting.\n", error.String()));
		#endif
		be_app->PostMessage(B_QUIT_REQUESTED);
		release_sem(mConstructor_sem);
		delete_sem(mConstructor_sem);
		return;
	} else {
		resume_thread(mReceive_th);
		#if DEBUG
		PRINT(("Announcing my presence to the world...\n"));
		#endif
		BMessage hellomsg(DASPORTAL_HELLO);
		//this forces the Receiver thread function to start executing before the
		//constructor finishes.
		if (acquire_sem(mConstructor_sem) != B_OK) {
			be_app->PostMessage(B_QUIT_REQUESTED);
			return;
		}
		Broadcast(&hellomsg);
		threadlist = new BList(1);
		queue = new BList(1);
		BMessage updmsg(UPDATE_QUEUE);
		BMessenger *msgr = new BMessenger(NULL, this, NULL);
		mqchecker = new BMessageRunner(*msgr, &updmsg, 1000000, -1);
		delete msgr;
		release_sem(mConstructor_sem);
	}
	delete_sem(mConstructor_sem);
}


NetCentral::~NetCentral() {
	#if DEBUG
	PRINT(("NetCentral destructor\n"));
	#endif
	delete mqchecker;
	delete_sem(mFS_sem);
	delete_sem(mFR_sem);
	status_t stat;
	wait_for_thread(mReceive_th, &stat);
	wait_for_thread(mFRC_th, &stat);
	BMessage goodbye(DASPORTAL_GOODBYE);
	Broadcast(&goodbye);
	delete mBroadcast_ep;
	while (!queue->IsEmpty())
		delete (thmaster*) queue->RemoveItem(0L);
	delete queue;
	delete threadlist;
	delete_sem(mQueueList_sem);
	delete_sem(mThreadList_sem);
}


bool NetCentral::QuitRequested() {
	#if DEBUG
	PRINT(("NetCentral QuitRequested\n"));
	#endif
	quit = 1;
	if (!threadlist->IsEmpty())
		return false;
	return true;
}


void NetCentral::MessageReceived(BMessage *msg)  {
	switch(msg->what) {
		case DASPORTAL_GET_FILE: {
			BString host;
			BMessage content(*msg);
			content.what = DASPORTAL_FILE_REQUEST;
			if ((msg->HasString("host")) && (!msg->HasString("target")))
				msg->FindString("host", &host);
			else
				msg->FindString("target", &host);
			SendTo(&content, (char*) host.String());
			}
			break;
		case UPDATE_QUEUE: {
			/*
				this function does double duty: if the threadlist contains items when the
				app is being shutdown, the app should not shutdown until those threads are
				done. in addition, under normal pre-quit circumstances, this will look for
				items in the file transfer queue for things to be sent.
			*/
				if (acquire_sem_etc(mThreadList_sem,1,B_ABSOLUTE_TIMEOUT,250000) != B_NO_ERROR)
					return;
				if (acquire_sem_etc(mQueueList_sem,1,B_ABSOLUTE_TIMEOUT,250000) != B_NO_ERROR) {
					release_sem(mThreadList_sem);
					return;
				}
				thmaster *tmp;
				status_t stat;
				thread_info ti;
				for (int32 i = 0; i < threadlist->CountItems() ; i++) {
					tmp = (thmaster*) threadlist->ItemAt(i);
					if ((stat = get_thread_info(tmp->thread, &ti)) == B_BAD_THREAD_ID) {
						#if DEBUG
						PRINT(("Dead thread detected\n"));
						#endif
						//Dead threads are those that have completed, but haven't removed themselves
						//from the queue for some reason.
						threadlist->RemoveItem(tmp);
						delete tmp->msg;
						delete tmp;
						continue;
					}
					#if DEBUG
					switch(ti.state) {
						case B_THREAD_RUNNING:
							{
								PRINT(("Thread is B_THREAD_RUNNING\n"));
							}break;
						case B_THREAD_READY:
							{
								PRINT(("Thread is B_THREAD_READY\n"));
							}break;
						case B_THREAD_WAITING:
							{
								PRINT(("Thread is B_THREAD_WAITING\n"));
							}break;
						case B_THREAD_RECEIVING:
							{
								PRINT(("Thread is B_THREAD_RECEIVING\n"));
							}break;
						case B_THREAD_ASLEEP:
							{
								PRINT(("Thread is B_THREAD_ASLEEP\n"));
							}break;
						case B_THREAD_SUSPENDED:
							{
								PRINT(("Thread is B_THREAD_SUSPENDED\n"));
							}break;
						default:
							PRINT(("Thread is in an unknown state!!\n"));
					}
					#endif
					if (ti.state == B_THREAD_SUSPENDED) { //our send threads should never be suspended, but...
						#if DEBUG
						PRINT(("Detected a suspended thread!\n"));
						#endif
						resume_thread(tmp->thread);
					}
				}
			if (!quit) {
				if (fscount < 5 ) {
					thmaster *tmp;
					if (!queue->IsEmpty())
						for (int i = 0; i < (5 - threadlist->CountItems()) ; i++) {
							atomic_add(&fscount, 1);
							tmp = (thmaster*) queue->RemoveItem(0L);
							if (tmp == NULL)
								continue;
							threadlist->AddItem(tmp);
							tmp->thread = spawn_thread(SpawnFileSender, "FileSendingThread", B_LOW_PRIORITY, tmp);
							resume_thread(tmp->thread);
						}
				}
			} else {
				if (threadlist->IsEmpty()) {
					be_app->PostMessage(B_QUIT_REQUESTED);
				} else {
				#if DEBUG
				PRINT(("Some (%ld) [send] file transfers are still in progress. Will quit when they have completed.\n", threadlist->CountItems()));
				#endif
				}
			}
			release_sem(mQueueList_sem);
			release_sem(mThreadList_sem);
			}
			break;
		case DASPORTAL_FILE_REQUEST: {
				thmaster *thst = new thmaster;
				thst->msg = new BMessage(*msg);
				if (fscount < 5 ) {
					atomic_add(&fscount, 1);
					acquire_sem(mThreadList_sem);
					threadlist->AddItem(thst);
					thst->thread = spawn_thread(SpawnFileSender, "FileSendingThread", B_LOW_PRIORITY, thst);
					resume_thread(thst->thread);
					release_sem(mThreadList_sem);
				} else {
					acquire_sem(mQueueList_sem);
					queue->AddItem(thst);
					release_sem(mQueueList_sem);
				}
			}
			break;
		case DASPORTAL_SEND_TO: {
			BString host;
			BMessage content;
			int32 count = 0;
			type_code tcode;
			msg->FindMessage("message", &content);
			msg->GetInfo("target",&tcode,&count);
			uint16 port = 0;
			for (int32 i = 0; i < count ; i++) {
				msg->FindString("target", i, &host);
				msg->FindInt16("port", (int16*) &port);
				SendTo(&content,(char*)  host.String(), port);
			}
			}
			break;
		case DASPORTAL_HELLO: {
				BMessage reply(DASPORTAL_HELLO_REPLY);
				BMessage shared;
				SettingsMsg->FindMessage("shared_files", &shared);
				reply.AddMessage("shared_files", &shared);
				BString target;
				uint16 port = 0;
				msg->FindString("host", &target);
				msg->FindInt16("port", (int16*) &port);
				SendTo(&reply, (char*) target.String(), port);
			}
			break;
		case DASPORTAL_HELLO_REPLY: {
				if (MainWindow) {
					BMessage addhost(*msg);
					addhost.what = DASPORTAL_ADD_HOST;
					BMessenger *msgr = new BMessenger(NULL, MainWindow, NULL);
					msgr->SendMessage(&addhost);
					delete msgr;
				}
				BMessage *reply = new BMessage(DASPORTAL_CONNECTED);
				BMessage shared;
				SettingsMsg->FindMessage("shared_files",  &shared);
				reply->AddMessage("shared_files",&shared);
				BString target;
				uint16 port;
				msg->FindString("host", &target);
				msg->FindInt16("port", (int16*) &port);
				SendTo(reply, (char*) target.String(), port);
				delete reply;
			}
			break;
		case DASPORTAL_FILE_LIST: {
				if (MainWindow) {
					BMessage *filelist = new BMessage(*msg);
					BMessenger *msgr = new BMessenger(NULL, MainWindow, NULL);
					msgr->SendMessage(filelist);
					delete msgr;
					delete filelist;
				}
			}
			break;
		case DASPORTAL_CONNECTED: {
				if (MainWindow) {
					BMessage *addhost = new BMessage(*msg);
					addhost->what = DASPORTAL_ADD_HOST;
					BMessenger *msgr = new BMessenger(NULL, MainWindow, NULL);
					msgr->SendMessage(addhost);
					delete msgr;
					delete addhost;
				}
			}
			break;
		case DASPORTAL_GOODBYE: {
				if (MainWindow) {
					BMessage *removehost = new BMessage(*msg);
					removehost->what = DASPORTAL_REMOVE_HOST;
					BMessenger *msgr = new BMessenger(NULL, MainWindow, NULL);
					msgr->SendMessage(removehost);
					delete msgr;
					delete removehost;
				}
			}
			break;
		default:
			BLooper::MessageReceived(msg);
			break;
	}
}


int32 NetCentral::SendTo(BMessage *msg, const char *host, uint16 port) {
	#if DEBUG
	PRINT(("NetCentral SendTo\n"));
	PRINT(("Message being sent to %s:\n", host));
	msg->PrintToStream();
	#endif
	BNetAddress addr(host, port);
	BNetBuffer buf;
	buf.AppendMessage(*msg);
	BNetEndpoint ep(SOCK_DGRAM);
	return (ep.SendTo(buf, addr));
}


int32 NetCentral::Broadcast(BMessage *msg, uint16 port) {
	#if DEBUG
	PRINT(("NetCentral Broadcast\n"));
	PRINT(("Message being sent to INADDR_BROADCAST:\n"));
	msg->PrintToStream();
	#endif
	BNetAddress addr(INADDR_BROADCAST, port);
	BNetBuffer buf;
	buf.AppendMessage(*msg);
	return (mBroadcast_ep->SendTo(buf, addr));
}


int32 NetCentral::SpawnReceiverThread(void *arg) {
	#if DEBUG
	PRINT(("NetCentral SpawnReceiverThread\n"));
	#endif
	NetCentral *obj = (NetCentral*)arg;
	return (obj->ReceiverThread());
}


int32 NetCentral::ReceiverThread() {
	#if DEBUG
	PRINT(("NetCentral ReceiverThread\n"));
	#endif
	BNetEndpoint *main_ep = NULL;
	int option=1;
	do {
		snooze(250000);
		if (main_ep!=NULL)
			delete main_ep;
		main_ep = new BNetEndpoint(SOCK_DGRAM);
		main_ep->SetOption(SO_BROADCAST, SOL_SOCKET, &option, sizeof(int));
		main_ep->SetNonBlocking(true);
		main_ep->SetTimeout(0x186a0);
		main_ep->Bind(DASPORTAL_DEFAULT_BCAST_PORT);
		if (quit)
			break;
	} while(main_ep->InitCheck() != B_OK);
	release_sem(mConstructor_sem);
	BMessenger *msgr = new BMessenger(NULL, this, NULL);
	hostent *hstent;
	sockaddr_in src;
	char host[0xff];
	uint16 port;
	while(!quit) {
		if (main_ep->IsDataPending(0x186a0)) {
			#if DEBUG
			PRINT(("NetCentral ReceiverThread: Data waiting\n"));
			#endif
			BNetBuffer *buffer= new BNetBuffer(0x186a0);
			BNetAddress addr;
			BMessage *message = new BMessage;
			memset(host, 0, 0xff);
			port=DASPORTAL_DEFAULT_BCAST_PORT;
			main_ep->ReceiveFrom(*buffer, 0x186a0, addr, 0);
			buffer->RemoveMessage(*message);
			delete buffer;
			addr.GetAddr(host, &port);
			addr.GetAddr(src);
			hstent=gethostbyaddr( (char*) &src.sin_addr, sizeof(src.sin_addr), AF_INET);
			if (hstent != NULL)
				if (hstent->h_name != NULL)
					strcpy(host,hstent->h_name);
			#if DEBUG
			PRINT(("Data is from %s:%d\n", host, port));
			#endif
			if (message->HasString("host"))
				message->ReplaceString("host", host);
			else
				message->AddString("host", host);
			//sample code hack to make things work for now: if stealing this code, do stuff better!
			port = DASPORTAL_DEFAULT_BCAST_PORT;
			message->AddInt16("port", port);
			msgr->SendMessage(message);
			delete message;
		} else {
			//If no data is waiting, try *not* to take as much processor time as possible.
			snooze(10000);
		}
	 }
	delete msgr;
	delete main_ep;
	exit_thread(B_OK);
	return B_OK;
}


int32 NetCentral::SpawnFileSender(thmaster *mystuff) {
	#if DEBUG
	PRINT(("NetCentral SpawnFileSender\n"));
	#endif
	return (ME->FileSenderThread(mystuff));
}


int32 NetCentral::FileSenderThread(thmaster *mystuff) {
	if (acquire_sem(mFS_sem) != B_NO_ERROR) {
		acquire_sem(mThreadList_sem);
		threadlist->RemoveItem(mystuff);
		release_sem(mThreadList_sem);
		atomic_add(&fscount, -1);
		queue->AddItem(mystuff);
		exit_thread(B_ERROR);
		return B_ERROR;
	}
	#if DEBUG
	PRINT(("FileSenderThread\n"));
	#endif
	BString host;
	uint16 port;
	mystuff->msg->FindString("host", &host);
	port = DASPORTAL_DEFAULT_BCAST_PORT + 0xa;
	BNetEndpoint *ep;
	ep = new BNetEndpoint(SOCK_STREAM);
	ep->SetNonBlocking(false);
//	ep->SetTimeout(1000000);
	status_t stat;
	if ((stat = ep->Connect(host.String(), port)) == B_OK) {
		#if DEBUG
		PRINT(("Connection established.\n"));
		#endif
		BNetBuffer *buf = new BNetBuffer;
		BMessage xfer(DASPORTAL_HELLO);
		buf->AppendMessage(xfer);
		ep->Send(*buf ,0);
		delete buf;
		buf = new BNetBuffer;
		ep->Receive(*buf, 0x186a0, 0);
		buf->RemoveMessage(xfer);
		delete buf;
		if(xfer.what == B_OK) {
			#if DEBUG
			PRINT(("fst, Digging up file details\n"));
			#endif
			xfer.MakeEmpty();
			xfer = *mystuff->msg;
			xfer.what = FILE_INFO;
			entry_ref ref;
			mystuff->msg->FindRef("refs", &ref);
			off_t size;
			int32 blocks = 0;
			BFile *file = new BFile(&ref, B_READ_ONLY);
			file->GetSize(&size);
			xfer.AddInt64("file_size", size);
			blocks = (size / 0x10000) + ((size % 0x10000) ? 1 : 0);
			xfer.AddInt32("block_count", blocks);
			BNodeInfo *ni = new BNodeInfo(file);
			char mime[B_MIME_TYPE_LENGTH];
			ni->GetType(mime);
			xfer.AddString("mime_type", mime);
			buf = new BNetBuffer;
			buf->AppendMessage(xfer);
			xfer.MakeEmpty();
			ep->Send(*buf, 0);
			delete buf;
			buf = new BNetBuffer;
			ep->Receive(*buf, 0x186a0);
			buf->RemoveMessage(xfer);
			delete buf;
			buf=NULL;
			if (xfer.what == B_OK) {
				int32 count = 0;
				int32 status = 0;
				ssize_t bytes_read;
				uint32 checksum;
				int32 closer=0;
				unsigned char *data = (unsigned char*) malloc (0x10000);
				while ((ep->Error() == B_NO_ERROR) && (closer!=DASPORTAL_RECEIVER_DONE)) {
					if (count < blocks) {
						if (buf != NULL)
							delete buf;
						buf = new BNetBuffer(0x10000);
						bytes_read = 0;
						memset(data, 0, 0x10000);
						bytes_read = file->ReadAt((count * 0x10000), data, 0x10000);
						checksum = 0;
						for (int32 i = 0; i < bytes_read; i++)
							checksum += *(data + i);
						PRINT(("Checksum for this block: %x\n",checksum));
						status = B_ERROR;
					} else {
						ep->Receive(&closer, sizeof(closer), 0);
						if (closer == DASPORTAL_RECEIVER_DONE)
							break;
						count--;
					}
					while ((status != B_OK) && (ep->Error() == B_NO_ERROR)) {
						PRINT(("Sending block size: %lu bytes\n", bytes_read));
						ep->Send(&bytes_read, sizeof(bytes_read), 0);
						PRINT(("Waiting for good response\n"));
						ep->Receive(&status, sizeof(status), 0);
						if (status == DASPORTAL_RECEIVER_DONE) {
							closer = status;
							break;
						}
						snooze(10000);
					}
					if (ep->Error() != B_NO_ERROR)
						break;
					PRINT(("good response received\n"));
					if (status != B_OK)
						break;
					ep->Send(&checksum, sizeof(checksum), 0);
					ep->Receive(&status, sizeof(status), 0);
					buf->AppendData(data, bytes_read);
					do {
						PRINT(("sending data %ld of %ld\n", count+1, blocks));
						ep->Send(*buf, 0);
						ep->Receive(&status, sizeof(status), 0);
						PRINT(("receiver status: %s\n", (status==B_OK) ? "good":"bad"));
						if (status == DASPORTAL_RECEIVER_DONE) {
							closer = status;
							break;
						}
						if (status == DASPORTAL_RESTART)
							{
							ep->Send(&bytes_read, sizeof(bytes_read), 0);
							ep->Receive(&status, sizeof(status), 0);
							ep->Send(&checksum, sizeof(checksum), 0);
							ep->Receive(&status, sizeof(status), 0);
						}
						if (ep->Error() != B_NO_ERROR)
							break;
					} while (status != B_OK);
					count++;
				}
				PRINT(("File transfer completed.\n"));
				delete buf;
				buf = NULL;
				free(data);
			} else {
				#if DEBUG
				PRINT(("receiving side has aborted file transfer\n"));
				#endif
			}
			delete ni;
			delete file;
		} else {
			acquire_sem(mThreadList_sem);
			threadlist->RemoveItem(mystuff);
			release_sem(mThreadList_sem);
			queue->AddItem(mystuff);
			ep->Close();
			delete ep;
			release_sem(mFS_sem);
			atomic_add(&fscount, -1);
			exit_thread(B_ERROR);
			return B_ERROR;
		}
		PRINT(("Closing connection\n"));
		ep->Close();
	} else {
		#if DEBUG
		PRINT(("Connection failed: %ld\n", stat));
		#endif
		acquire_sem(mThreadList_sem);
		threadlist->RemoveItem(mystuff);
		release_sem(mThreadList_sem);
		queue->AddItem(mystuff);
		delete ep;
		release_sem(mFS_sem);
		atomic_add(&fscount, -1);
		exit_thread(B_ERROR);
		return B_ERROR;
	}
	delete ep;
	acquire_sem(mThreadList_sem);
	threadlist->RemoveItem(mystuff);
	release_sem(mThreadList_sem);
	if (mystuff != NULL) {
		delete mystuff->msg;
		delete mystuff;
	}
	release_sem(mFS_sem);
	atomic_add(&fscount, -1);
	exit_thread(B_OK);
	return B_OK;
}


int32 NetCentral::SpawnFileReceiverThread(BNetEndpoint *ep) {
	#if DEBUG
	PRINT(("NetCentral SpawnFileReceiverThread\n"));
	#endif
	return (ME->FileReceiverThread(ep));
}


int32 NetCentral::FileReceiverThread(BNetEndpoint *ep) {
	#if DEBUG
	PRINT(("NetCentral FileReceiverThread\n"));
	#endif
	if (acquire_sem(mFR_sem) != B_OK) {
		ep->Close();
		delete ep;
		atomic_add(&frcount, -1);
		exit_thread(B_ERROR);
		return B_ERROR;
	}
	BNetAddress addr = ep->RemoteAddr();
	hostent *hstent;
	sockaddr_in src;
	uint16 port;
	char host[255];
	addr.GetAddr(host, &port);
	addr.GetAddr(src);
	hstent = gethostbyaddr( (char*) &src.sin_addr, sizeof(src.sin_addr), AF_INET);
	if (hstent != NULL)
		if (hstent->h_name != NULL)
			strcpy(host, hstent->h_name);
	#if DEBUG
	PRINT(("Connection is from %s\n", host));
	#endif
	BMessage xfer;
	BNetBuffer *buf = new BNetBuffer;
	ep->SetNonBlocking(false);
	ep->Receive(*buf,0x186a0);
	buf->RemoveMessage(xfer);
	xfer.MakeEmpty();
	xfer.what = B_OK;
	delete buf;
	buf = new BNetBuffer;
	buf->AppendMessage(xfer);
	ep->Send(*buf, 0);
	delete buf;
	buf = new BNetBuffer;
	if (ep->Receive(*buf, 0x186a0) > -1) {
		xfer.MakeEmpty();
		xfer.what = 0;
		buf->RemoveMessage(xfer);
		if (xfer.what == FILE_INFO) {
			BString fname;
			entry_ref ref;
			SettingsMsg->FindRef("DownloadDirectory", &ref);
			BFile *file = new BFile;;
			BString mime;
			int32 blocks = 0;
			xfer.FindString("filename", &fname);
			BPath path(&ref);
			path.Append(fname.String());
			#if DEBUG
			PRINT(("file will be downloaded as:\n\t%s\n", path.Path()));
			#endif
			if (file->SetTo(path.Path(), B_WRITE_ONLY | B_CREATE_FILE | B_ERASE_FILE) == B_OK) {
				xfer.FindString("mime_type", &mime);
				BNodeInfo ni(file);
				ni.SetType(mime.String());
				off_t size = 0, bytes_recv = 0L;
				xfer.FindInt32("block_count", &blocks);
				xfer.FindInt64("file_size", &size);
				PRINT(("file size: %Ld bytes\n",size));
				xfer.MakeEmpty();
				xfer.what = B_OK;
				delete buf;
				buf = new BNetBuffer;
				buf->AppendMessage(xfer);
				ep->Send(*buf, 0);
				delete buf;
				buf = NULL;
				bytes_recv = 0;
				unsigned char *data = (unsigned char *) malloc (0x10000), tmp = '\0';
				size_t blocksize = 0, sz = 0;
				int32 bytes_rec = 0, count = 0, status = 0;
				uint32 expectedchecksum = 0, checksum = 0, block = 0;
				while ((ep->Error() == B_NO_ERROR) && (bytes_recv < size)) {
					if (buf!=NULL)
						delete buf;
					buf = new BNetBuffer(0x10000);
					memset(data, 0, 0x10000);
					blocksize = 0;
					while(((ep->Receive(&blocksize, sizeof(blocksize), 0)) != sizeof(blocksize)) &&
						(ep->Error() == B_NO_ERROR)) {
						PRINT(("Invalid blocksize read from network\n", blocksize));
						status=DASPORTAL_RESTART;
						ep->Send(&status, sizeof(status), 0);
						snooze(10000);
					}
					if (ep->Error() != B_NO_ERROR)
						break;
					if ((blocksize > 0x10000) || (blocksize == 0))
						continue;
					status = B_OK;
					ep->Send(&status, sizeof(status), 0);
					expectedchecksum = 0;
					ep->Receive(&expectedchecksum, sizeof(expectedchecksum), 0);
					ep->Send(&status, sizeof(status), 0);
					PRINT(("expecting %ld bytes, checksum: %x\n", blocksize, expectedchecksum));
					sz = 0;
					count = 0;
					do {
						bytes_rec = 0;
						bytes_rec = ep->Receive(*buf, blocksize - sz, 0);
						if (bytes_rec == 0) {
							count++;
							if (count >= 0xa)
								break;
							continue;
						}
						else
							count = 0;
						if (ep->Error() == B_NO_ERROR)
							sz += bytes_rec;
						PRINT(("*** start\nblock size: %lu\nbytes received: %ld\ntotal received %lu\n*** stop\n",
							blocksize, bytes_rec, sz));
					}while((ep->Error() == B_NO_ERROR) && (sz < blocksize));
					if (ep->Error() == B_NO_ERROR) {
						if (sz != blocksize) {
							PRINT(("Incorrect amount of data received\n"));
							status = B_ERROR;
							ep->Send(&status, sizeof(status), 0);	
							continue;
						}
					} else
						break;
					checksum=0;
					PRINT(("Received block (%lu/%lu)\n", sz, blocksize));
					buf->RemoveData(data, sz);
					for (uint32 i = 0; i < sz; i++) {
						tmp = '\0';
						tmp = *(data + i);
						checksum += tmp;
					}
					PRINT(("actual checksum: %x\n", checksum));
					if (expectedchecksum != checksum) {
						PRINT(("Checksum mismatch\n\tis: %x\n\tshould be: %x\n", checksum, expectedchecksum));
						status = B_ERROR;
						ep->Send(&status, sizeof(status), 0);
						continue;
					}
					file->Write(data, sz);
					block++;
					status = B_OK;
					ep->Send(&status, sizeof(status), 0);
					bytes_recv += sz;
					}
				int32 closer = DASPORTAL_RECEIVER_DONE;
				ep->Send(&closer, sizeof(closer), 0);
				if (bytes_recv == size)
					PRINT(("File transfer successful.\n"));
				else
					PRINT(("File transfer failed (%Ld/%Ld)\n", bytes_recv, size));
				free(data);
				delete buf;
				buf = NULL;
			} else {
				#if DEBUG
				PRINT(("File transfer aborted on receiver side.\n"));
				#endif
				xfer.MakeEmpty();
				xfer.what = FILE_ABORT;
				delete buf;
				buf->AppendMessage(xfer);
				ep->Send(buf ,0);
			}
			delete file;
		}
	}
	if (buf != NULL)
		delete buf;
	ep->Close();
	delete ep;
	release_sem(mFR_sem);
	atomic_add(&frcount, -1);
	exit_thread(B_OK);
	return B_OK;
}


int32 NetCentral::SpawnFRConnection(void *arg) {
	#if DEBUG
	PRINT(("NetCentral SpawnFRConnection\n"));
	#endif
	return (ME->FRConnection());
}


int32 NetCentral::FRConnection() {
	#if DEBUG
	PRINT(("NetCentral FRConnection\n"));
	#endif
	BNetEndpoint *ep = new BNetEndpoint(SOCK_STREAM), *tmp = NULL;
	ep->SetNonBlocking(true);
	ep->Bind(DASPORTAL_DEFAULT_BCAST_PORT + 0xa);
	ep->Listen(5);
	ep->SetTimeout(0x186a0);
	while(!quit) {
		tmp = NULL;
		tmp = ep->Accept(10000);
		if (tmp != NULL) {
			#if DEBUG
			PRINT(("Inbound connection!\n"));
			#endif
			if (frcount < 5) {
				atomic_add(&frcount, 1);
				resume_thread(spawn_thread(SpawnFileReceiverThread, "FileReceivingThread", B_LOW_PRIORITY, tmp));
			} else {
				BNetBuffer *buf=new BNetBuffer;
				tmp->Receive(*buf, 10000, 0);
				delete buf;
				buf = new BNetBuffer;
				BMessage goaway(FILE_TRY_LATER);
				buf->AppendMessage(goaway);
				tmp->Send(*buf, 0);
				tmp->Close();
				delete tmp;
				delete buf;
			}
		} else {
			snooze(10000);
		}
	}
	ep->Close();
	delete ep;
	exit_thread(B_OK);
	return B_OK;
}
